"""
informant/entry.py

This module contains entry (feed item) related functions.
"""

from informant.config import InformantConfig
import informant.ui as ui

class Entry:
    def __init__(self, title, timestamp, body, feed_name):
        self.title = title
        self.timestamp = timestamp
        self.pretty_date = timestamp.strftime('%a, %d %b %Y %H:%M:%S %z')
        self.body = body
        self.feed_name = feed_name

    def has_been_read(self):
        """ Check if this entry has been read and return True or False. """
        readlist = InformantConfig().readlist
        #ui.debug_print('readlist: {}'.format(readlist))
        title = self.title
        date = self.timestamp
        if str(date.timestamp()) + '|' + title in readlist:
            return True
        return False

    def mark_as_read(self):
        """ Save this entry to mark it as read. """
        readlist = InformantConfig().readlist
        if self.has_been_read():
            return
        title = self.title
        date = self.timestamp
        readlist.append(str(date.timestamp()) + '|' + title)




"""
informant/config.py

This module contains helpers to manage arguments, options and configuration
settings provided to Informant.
"""

import json
import os

DEBUG_OPT = '--debug'
FILE_OPT = '--file'
CFILE_OPT = '--config'
NOCACHE_OPT = '--no-cache'
CLRCACHE_OPT = '--clear-cache'
CLRREAD_OPT = '--clear-readlist'
PAGER_OPT = '--pager'

FILE_DEFAULT = '/var/lib/informant.dat' # readlist save file
CACHE_DEFAULT = '/var/cache/informant' # http caching
CONFIG_BASE = 'informantrc.json' # user config
PAGER_DEFAULT = os.environ.get('INFORMANT_PAGER', default=None)

class Singleton(type):
    """ A Singleton class to be used as a base """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class InformantConfig(metaclass=Singleton):
    """ Class to store global Informant arguments/options and configuration """

    def __init__(self):
        self.argv = {}
        self.config = None
        self.colors = {
                'RED': '\033[0;31m',
                'YELLOW': '\033[1;33m',
                'CLEAR': '\033[0m',
                'BOLD': '\033[1m'
        }
        self.readlist = None
        self.debug_print = None

    def set_argv(self, args):
        self.argv = args

    def get_argv(self):
        return self.argv

    def get_argv_debug(self):
        return self.argv.get(DEBUG_OPT)

    def get_argv_use_cache(self):
        """ Return True if we should use the cache, else False.
        Providing the NOCACHE_OPT means that we should not use the cache.
        """
        if self.argv.get(NOCACHE_OPT):
            return False
        return True

    def get_argv_clear_cache(self):
        """ Return True if we should clear the cache, else False.
        Providing the CLRCACHE_OPT means we want to clear the cache.
        """
        if self.argv.get(CLRCACHE_OPT):
            return True
        return False

    def get_cachefile(self):
        return CACHE_DEFAULT

    def get_argv_clear_savefile(self):
        """ Return True if we should clear the savefile (readlist), else False.
        Providing the CLRREAD_OPT means we want to clear the savefile.
        """
        if self.argv.get(CLRREAD_OPT):
            return True
        return False


    def get_savefile(self):
        if self.argv.get(FILE_OPT):
            return self.argv.get(FILE_OPT)
        return FILE_DEFAULT

    def set_config(self, config):
        self.config = config

    def get_pager(self):
        if self.argv.get(PAGER_OPT):
            return self.argv.get(PAGER_OPT)
        return PAGER_DEFAULT

    def read_config(self):
        self.config = {}
        cfile_option = self.argv.get(CFILE_OPT)
        cfg_fname = None
        if cfile_option and os.path.exists(cfile_option):
            cfg_fname = cfile_option
        elif os.path.exists(os.path.expandvars('$HOME/.' + CONFIG_BASE)):
            cfg_fname = os.path.expandvars('$HOME/.' + CONFIG_BASE)
        elif os.path.exists(os.path.expandvars('$XDG_CONFIG_HOME/' + CONFIG_BASE)):
            cfg_fname = os.path.expandvars('$XDG_CONFIG_HOME/' + CONFIG_BASE)
        elif os.path.exists(os.path.join('/etc/', CONFIG_BASE)):
            cfg_fname = os.path.join('/etc/', CONFIG_BASE)
        else:
            xdg_config_dirs = os.environ.get('$XDG_CONFIG_DIRS')
            if xdg_config_dirs:
                xdg_config_dirs = [xdg_config_dirs.split(':')]
                for dirname in xdg_config_dirs:
                    if os.path.exists(os.path.join(dirname, CONFIG_BASE)):
                        cfg_fname = os.path.join(dirname, CONFIG_BASE)
                        break
        if self.get_argv_debug():
            self.debug_print('cfg_fname: {}'.format(cfg_fname))
        if cfg_fname is not None:
            with open(cfg_fname, 'r') as cfg:
                self.config = json.loads(cfg.read())
        if self.get_argv_debug():
            self.debug_print('config: {}'.format(self.config))
        return self.config

    def get_config(self):
        if self.config is None:
            return self.read_config()
        return self.config



"""
informant/ui.py

This module contains User Interface related functions.
"""

import os
import shutil
import sys
import textwrap

import html2text
import psutil

from informant.config import PAGER_DEFAULT, InformantConfig

RAW_OPT = '--raw'

def err_print(*args, **kwargs):
    """ Same as builtin print but output to stderr with red color and "ERROR"
    preamble.
    """
    red = InformantConfig().colors['RED']
    clear = InformantConfig().colors['CLEAR']
    msg = red + 'ERROR: ' + clear
    for arg in args:
        msg += arg
    print(msg, file=sys.stderr, **kwargs)

def warn_print(*args, **kwargs):
    """ Same as builtin print but output to stderr with yellow color and
    "WARNING" preamble.
    """
    yellow = InformantConfig().colors['YELLOW']
    clear = InformantConfig().colors['CLEAR']
    msg = yellow + 'WARN: ' + clear
    for arg in args:
        msg += arg
    print(msg, file=sys.stderr, **kwargs)

def debug_print(*args, **kwargs):
    """ Same as builtin print but output to stderr and only when the debug
    option is provided.
    """
    if InformantConfig().get_argv_debug():
        print(*args, file=sys.stderr, **kwargs)

def pacman_msg(*args, **kwargs):
    """ Same as print but include yellow color and "informant" preamble so the
    message is clear in pacman.
    """
    yellow = InformantConfig().colors['YELLOW']
    clear = InformantConfig().colors['CLEAR']
    msg = yellow + ':: informant: ' + clear
    for arg in args:
        msg += arg
    print(msg, **kwargs)

def prompt_yes_no(question, default):
    """ Print 'question' to user with (y/n) and 'default' being the
    response for blank input.
    """
    again = 'Unknown response.'
    if default.lower() in ('y', 'yes'):
        options = '(Y/n): '
    elif default.lower() in ('n', 'no'):
        options = '(y/N): '

    response = input(' '.join((question, options))).lower()
    while response not in ('y', 'yes', 'n', 'no', ''):
        response = input(' '.join((again, question, options))).lower()
    if response == '':
        return default
    return response

def running_from_pacman():
    """ Return True if the parent process is pacman """
    ppid = os.getppid()
    p_name = psutil.Process(ppid).name()
    debug_print('informant running from: {}'.format(p_name))
    return p_name == 'pacman'

def format_content(entry, body, timestamp, title) -> str:
    if entry.feed_name is not None:
        feed_name = '({})'.format(entry.feed_name)
        content = '{}\n{}\n{}\n\n{}'.format(title, feed_name, timestamp, body)
    else:
        content = '{}\n{}\n\n{}'.format(title, timestamp, body)
    return content

def format_body(body) :
    h2t = html2text.HTML2Text()
    h2t.inline_links = False
    h2t.body_width = 85
    body = h2t.handle(body)
    return body

def pretty_print_item(entry):
    """ Print out the given entry, replacing some markup to make it look nicer.
    If the '--raw' option has been provided then the markup will not be
    replaced. """
    argv = InformantConfig().get_argv()
    title = entry.title
    body = entry.body
    timestamp = str(entry.pretty_date)
    pager = InformantConfig().get_pager()
    pager_is_valid = isinstance(pager, str) and shutil.which(pager) is not None
    bold = InformantConfig().colors['BOLD']
    clear = InformantConfig().colors['CLEAR']
    if not argv.get(RAW_OPT):
        body = format_body(body)              
        if pager_is_valid:
            # format the title as a markdown header
            # for a pager.
            title = f"# {title}"
        else:
            #if not using raw also bold title
            title = bold + title + clear         
    
    if pager_is_valid:
        with os.popen(pager, 'w') as pipe:
            pipe.write(format_content(entry, body, timestamp, title))
    else:
        print(format_content(entry, body, timestamp, title))

def format_list_item(entry, index):
    """ Returns a formatted string with the entry's index number, title, and
    right-aligned timestamp. Unread items are bolded"""
    bold = InformantConfig().colors['BOLD']
    clear = InformantConfig().colors['CLEAR']
    terminal_width = shutil.get_terminal_size().columns
    timestamp = str(entry.pretty_date)
    wrap_width = terminal_width - len(timestamp) - 1
    heading = str(index) + ': ' + entry.title
    wrapped_heading = textwrap.wrap(heading, wrap_width)
    padding = terminal_width - len(wrapped_heading[0] + timestamp)
    if entry.has_been_read():
        return (
            wrapped_heading[0] +
            ' ' * (padding) +
            timestamp +
            '\n'.join(wrapped_heading[1:])
                )
    else:
        return (
            bold +
            wrapped_heading[0] +
            clear +
            ' ' * (padding) +
            timestamp +
            bold +
            '\n'.join(wrapped_heading[1:]) +
            clear
        )




"""
informant/file.py

This module contains filesystem related functions.
"""

import glob
import os
import pickle
import shutil
import sys

from informant.config import InformantConfig
import informant.ui as ui

def read_datfile():
    """ Return the saved readlist from the datfile """
    filename = InformantConfig().get_savefile()
    ui.debug_print('Getting datfile from "{}"'.format(filename))
    if InformantConfig().get_argv_clear_savefile():
        ui.debug_print('Clear savefile specified returning empty list')
        return []
    try:
        with open(filename, 'rb') as pickle_file:
            try:
                readlist = pickle.load(pickle_file)
                pickle_file.close()
                if isinstance(readlist, tuple):
                    # backwards compatibility with informant < 0.4.0 save data
                    readlist = readlist[1]
            except (EOFError, ValueError):
                readlist = []
    except (FileNotFoundError, PermissionError):
        readlist = []
    return readlist

def save_datfile():
    """ Save the readlist to the datfile """
    debug = InformantConfig().get_argv_debug()
    readlist = InformantConfig().readlist
    if debug:
        ui.debug_print('running in debug mode, will not update readlist')
        return
    filename = InformantConfig().get_savefile()
    try:
        # then open as write to save updated list
        with open(filename, 'wb') as pickle_file:
            pickle.dump(readlist, pickle_file)
            pickle_file.close()
    except PermissionError:
        ui.err_print('Unable to save read information, please re-run with \
correct permissions to access "{}".'.format(filename))
        sys.exit(255) # this should never block pacman because the hook should run with root/sudo

def clear_cachefile():
    """ Empty the cachefile directory """
    cache_dir = InformantConfig().get_cachefile()
    pattern = os.path.join(cache_dir, '*')
    ui.debug_print('Removing based on pattern: {}'.format(pattern))
    for filename in glob.glob(pattern):
        if os.path.isdir(filename):
            shutil.rmtree(filename)
        else:
            os.remove(filename)




#!/usr/bin/python3
"""
informant - an Arch Linux News reader designed to also be used as a pacman hook

Usage:
    informant [options] check
    informant [options] list [--reverse --unread]
    informant [options] read [<item> | --all]

Commands:
    check - Check for unread news items, will exit with a positive return code
            with the number of unread items. If there is only one unread item it
            will also print in like 'read' would and mark it as read, else it
            will print the number of unread titles.

    list -  Print the most recent news items, regardless of read status. If
            '--reverse' is provided items are printed oldest to newest. If
            '--unread' is provided only the unread items are returned.

    read  - Read the specified news item, <item> can be either an index or a
            full title. This will also save the item as 'read' so that future
            calls to 'check' will no longer display it. If no <item> is given,
            will begin looping through all unread items, printing each one and
            marking them as read with a prompt to continue. Passing the --all
            flag will mark all items as read without printing.

Options:
    -c <cfile>, --config=<cfile>    Use <cfile> as the config file
    -d, --debug                     Print the debug messages and don't make
                                    changes to the save file
    -r, --raw                       When printing items do not replace any
                                    markup
    -f <file>, --file=<file>        Use <file> as the save location for read
                                    items
    -p <pager>, --pager=<pager>     Use <pager> as the pager to display the
                                    news items
    --no-cache                      Do not use cache
    --clear-cache                   Empty the cache before fetching feed(s)
    --clear-readlist                Empty the saved readlist
    -h, --help                      Show this help and exit
    -V,--version                    Show version and exit

"""

# builtins
import sys

# external
import docopt

# local
from informant.config import InformantConfig
from informant.feed import Feed
import informant.file as fs
import informant.ui as ui

__version__ = '0.6.0'

# commands
CHECK_CMD = 'check'
LIST_CMD = 'list'
READ_CMD = 'read'

# 'list' options
REV_OPT = '--reverse'
UNREAD_OPT = '--unread'

# 'read' options and args
ITEM_ARG = '<item>'
READALL_OPT = '--all'

def check_cmd(feed):
    """ Run the check command. Check if there are any news items that are
    unread. If there is only one unread item, print it out and mark it as read.
    Also, exit the program with return code matching the unread count. """
    running_from_pacman = ui.running_from_pacman()
    unread = 0
    unread_items = []
    for entry in feed:
        if not entry.has_been_read():
            unread += 1
            unread_items.append(entry)
    if unread == 1:
        if running_from_pacman:
            ui.pacman_msg('Stopping upgrade to print news')
        ui.pretty_print_item(unread_items[0])
        unread_items[0].mark_as_read()
        fs.save_datfile()
        if running_from_pacman:
            ui.pacman_msg('You can re-run your pacman command to complete the upgrade')
    elif unread > 1:
        print('There are {:d} unread news items! Use informant to read \
them.'.format(unread))
        if running_from_pacman:
            ui.pacman_msg('Run `informant read` before re-running your pacman command')
    else:
        print('There are no unread news items')
    sys.exit(unread)

def list_cmd(feed):
    """ Run the list command. Print out a list of recent news item titles. """
    argv = InformantConfig().get_argv()
    if argv.get(REV_OPT):
        feed_list = reversed(feed)
    else:
        feed_list = feed
    index = 0
    for entry in feed_list:
        if not argv.get(UNREAD_OPT) \
        or (argv.get(UNREAD_OPT) and not entry.has_been_read()):
            print(ui.format_list_item(entry, index))
            index += 1

def read_cmd(feed):
    """ Run the read command. Print news items and mark them as read. """
    argv = InformantConfig().get_argv()
    if argv.get(READALL_OPT):
        for entry in feed:
            entry.mark_as_read()
    else:
        if argv[ITEM_ARG]:
            try:
                item = int(argv[ITEM_ARG])
                entry = feed[item]
            except ValueError:
                for entry in feed:
                    if entry.title == item:
                        break
                #NOTE: this will read the oldest unread item if no matches are found
            ui.pretty_print_item(entry)
            entry.mark_as_read()
        else:
            unread_entries = list()
            for entry in feed:
                if not entry.has_been_read():
                    unread_entries.insert(0, entry)
            for entry in unread_entries:
                ui.pretty_print_item(entry)
                entry.mark_as_read()
                if entry is not unread_entries[-1]:
                    read_next = ui.prompt_yes_no('Read next item?', 'yes')
                    if read_next in ('n', 'no'):
                        break
                else:
                    print('No more unread items')
    fs.save_datfile()

def main():
    """ The main function.
    Check given arguments get feed and run given command. """
    argv = docopt.docopt(__doc__, version='informant v{}'.format(__version__))
    InformantConfig().set_argv(argv)
    InformantConfig().debug_print = ui.debug_print
    InformantConfig().readlist = fs.read_datfile()
    config = InformantConfig().get_config()
    ui.debug_print('cli args: {}'.format(argv))

    if 'feeds' in config:
        feed = []
        for config_feed in config['feeds']:
            feed += Feed(config_feed).entries
    else:
        feed = Feed().entries

    if not feed:
        ui.warn_print('no news feed items, informant is performing no action')
        sys.exit()

    feed = sorted(feed, key=lambda k: k.timestamp, reverse=True)

    if argv.get(CHECK_CMD):
        check_cmd(feed)
    elif argv.get(LIST_CMD):
        list_cmd(feed)
    elif argv.get(READ_CMD):
        read_cmd(feed)
    sys.exit()

if __name__ == '__main__':
    main()



"""
informant/feed.py

This module defines the structure of a newsfeed for Informant.
"""

import os
import sys

import requests
import feedparser
from dateutil import parser as date_parser
from cachecontrol import CacheControl
from cachecontrol.caches import FileCache
from urllib.error import URLError

from informant.config import InformantConfig
from informant.entry import Entry
import informant.file as fs
import informant.ui as ui

ARCH_NEWS = 'https://archlinux.org/feeds/news/'

class Feed:
    def __init__(self, config={}):
        if 'name' in config:
            self.name = config['name']
        else:
            self.name = None

        if 'url' in config:
            self.url = config['url']
        else:
            self.url = ARCH_NEWS

        if 'title-key' in config:
            self.title_key = config['title-key']
        else:
            self.title_key = 'title'

        if 'body-key' in config:
            self.body_key = config['body-key']
        else:
            self.body_key = 'summary'

        if 'timestamp-key' in config:
            self.timestamp_key = config['timestamp-key']
        else:
            self.timestamp_key = 'published'

        ui.debug_print('building feed for: {}'.format(self.name if self.name is not None else self.url))

        self.feed = self.fetch()  # the complete feed as returned by feedparser
        self.entries = self.build_feed()  # the list of entries informant will use

    def build_feed(self):
        """
        Abstract away any differences in feeds by using the parsed keys and
        return an informant-friendly list of entries
        """
        entries = []
        for item in self.feed.entries:
            timestamp = date_parser.parse(item[self.timestamp_key])
            entries.append(Entry(item[self.title_key],
                                 timestamp,
                                 item[self.body_key],
                                 self.name))
        return entries

    def fetch(self):
        # TODO: update to check for http 301/302 redirects
        feed = None
        if InformantConfig().get_argv_clear_cache():
            ui.debug_print('Clearing cache')
            fs.clear_cachefile()
        if InformantConfig().get_argv_use_cache():
            ui.debug_print('Checking cache in {}'.format(InformantConfig().get_cachefile()))
            cachefile = InformantConfig().get_cachefile()
            os.umask(0o0002) # unrestrict umask so we can cache with proper permissions
            try:
                session = CacheControl(requests.Session(), cache=FileCache(cachefile, filemode=0o0664, dirmode=0o0775))
                feed = feedparser.parse(session.get(self.url).content)
            except Exception as e:
                ui.err_print('Unable to read cache information: {}'.format(e))
                ui.debug_print('Falling back to fetching feed')
                feed = feedparser.parse(self.url)
        else:
            feed = feedparser.parse(self.url)

        if feed.bozo:
            e = feed.bozo_exception
            if isinstance(e, URLError):
                # most likely this is an internet issue (no connection)
                ui.warn_print('News could not be fetched for {}'.format(self.name if self.name is not None else self.url))
                ui.debug_print('URLError: {}'.format(e.reason))
            else:
                # I think this is most likely to be a malformed feed
                ui.err_print('Encountered feed error: {}'.format(feed.bozo_exception))
                ui.debug_print('bozo message: {}'.format(feed.bozo_exception.getMessage()))
            # In either of these error cases we probably shouldn't return error
            # so the pacman hook won't hold up an operation.
            # Here return an empty set of entries in case only one of multiple
            # feeds failed to fetch
            try:
                feed = feedparser.util.FeedParserDict()
                feed.update({'entries': []})
            except Exception as e:
                ui.err_print('Unexpected error: {}'.format(e))
                sys.exit()

        return feed
